package envoy.client.ui;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javafx.application.Application;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.stage.Stage;

import envoy.client.data.*;
import envoy.client.net.Client;
import envoy.client.net.WriteProxy;
import envoy.data.Message;
import envoy.data.User;
import envoy.data.User.UserStatus;
import envoy.exception.EnvoyException;
import envoy.util.EnvoyLog;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>Startup.java</strong><br>
 * Created: <strong>26.03.2020</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @since Envoy Client v0.1-beta
 */
public final class Startup extends Application {

	private LocalDB			localDB;
	private Client			client;
	private WriteProxy		writeProxy;
	private Cache<Message>	cache;

	private static final ClientConfig	config	= ClientConfig.getInstance();
	private static final Logger			logger	= EnvoyLog.getLogger(Startup.class);

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void start(Stage stage) throws Exception {
		try {
			// Load the configuration from client.properties first
			final Properties properties = new Properties();
			properties.load(Startup.class.getClassLoader().getResourceAsStream("client.properties"));
			config.load(properties);

			// Override configuration values with command line arguments
			final String[] args = getParameters().getRaw().toArray(new String[0]);
			if (args.length > 0) config.load(args);

			// Check if all mandatory configuration values have been initialized
			if (!config.isInitialized()) throw new EnvoyException("Configuration is not fully initialized");
		} catch (final Exception e) {
			new Alert(AlertType.ERROR, "Error loading configuration values:\n" + e);
			e.printStackTrace();
			System.exit(1);
		}

		// Setup logger for the envoy package
		EnvoyLog.initialize(config);
		EnvoyLog.attach("envoy");
		EnvoyLog.setFileLevelBarrier(config.getFileLevelBarrier());
		EnvoyLog.setConsoleLevelBarrier(config.getConsoleLevelBarrier());

		// Initialize the local database
		if (config.isIgnoreLocalDB()) {
			localDB = new TransientLocalDB();
			new Alert(AlertType.WARNING, "Ignoring local database.\nMessages will not be saved!").showAndWait();
		} else try {
			localDB = new PersistentLocalDB(new File(config.getHomeDirectory(), config.getLocalDB().getPath()));
		} catch (final IOException e3) {
			logger.log(Level.SEVERE, "Could not initialize local database", e3);
			new Alert(AlertType.ERROR, "Could not initialize local database!\n" + e3).showAndWait();
			System.exit(1);
			return;
		}

		// Initialize client and unread message cache
		client	= new Client();
		cache	= new Cache<>();

		// Try to connect to the server
		new LoginDialog(client, localDB, cache).showAndWait();

		// Set client user in local database
		localDB.setUser(client.getSender());

		// Initialize chats in local database
		try {
			localDB.initializeUserStorage();
			localDB.loadUserData();
		} catch (final FileNotFoundException e) {
			// The local database file has not yet been created, probably first login
		} catch (final Exception e) {
			e.printStackTrace();
			new Alert(AlertType.ERROR, "Error while loading local database: " + e + "\nChats will not be stored locally.").showAndWait();
		}

		// Initialize write proxy
		writeProxy = client.createWriteProxy(localDB);

		if (client.isOnline()) {

			// Save all users to the local database and flush cache
			localDB.setUsers(client.getUsers());
			writeProxy.flushCache();
		} else
			// Set all contacts to offline mode
			localDB.getUsers()
				.values()
				.stream()
				.filter(User.class::isInstance)
				.map(User.class::cast)
				.forEach(u -> u.setStatus(UserStatus.OFFLINE));

		// Prepare stage and load ChatScene
		final var sceneContext = new SceneContext(stage);
		sceneContext.load(SceneContext.SceneInfo.CHAT_SCENE);
		sceneContext.<ChatSceneController>getController().initializeData(sceneContext, localDB, client, writeProxy);

		stage.setTitle("Envoy");
		stage.setMinHeight(400);
		stage.setMinWidth(350);
		stage.getIcons().add(IconUtil.load("/icons/envoy_logo.png"));
		stage.show();

		// Relay unread messages from cache
		if (cache != null && client.isOnline()) cache.relay();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void stop() {
		try {
			// Save Settings and PersistentLocalDB on shutdown
			logger.info("Closing connection...");
			client.close();

			logger.info("Saving local database and settings...");
			localDB.save();
			Settings.getInstance().save();
		} catch (final Exception e) {
			logger.log(Level.SEVERE, "Unable to save local files", e);
		}
	}

	/**
	 * Starts the application.
	 *
	 * @param args the command line arguments are processed by the
	 *             {@link ClientConfig}
	 * @since Envoy Client v0.1-beta
	 */
	public static void main(String[] args) { launch(args); }
}
